import type { Listing } from "$lib/Listing";
import {
	MongoClient,
	ObjectId,
	ServerApiVersion,
	type Document,
	type WithId,
} from "mongodb";
import type { Id } from "../Id";

type ListingFromMongo = Omit<Listing, "id"> & { _id: ObjectId };

const uri = `mongodb://127.0.0.1:27017`;
const DATABASE = "ten-thousand";
let cachedClient: MongoClient | null = null;

async function getClient() {
	if (cachedClient !== null) return cachedClient;

	const c = new MongoClient(uri, {
		serverApi: {
			version: ServerApiVersion.v1,
			strict: true,
			deprecationErrors: true,
		},
	});

	await c.connect();
	return c;
}

export async function writeListing(col: string, listing: Listing) {
	const client = await getClient();

	await client.db(DATABASE).collection(col).insertOne(fixListingForMongo(listing));
}

export async function listCollection<T>(
	col: string,
	dataGuard: (target: unknown) => target is Listing<T>,
) {
	const client = await getClient();
	const res = await client.db(DATABASE).collection(col).find();

	return (await res.toArray()).map((doc) => fixListingFromMongo(doc, dataGuard));
}

export async function readListingById<T>(
	col: string,
	id: Id,
	dataGuard: (target: unknown) => target is Listing<T>,
) {
	const client = await getClient();
	const res = await client.db(DATABASE).collection(col).findOne({ _id: id });

	if (res === null) return null;
	return fixListingFromMongo(res, dataGuard);
}

export async function readListingByQuery<T>(
	col: string,
	query: object,
	dataGuard: (target: unknown) => target is Listing<T>,
): Promise<Listing<T> | null> {
	const client = await getClient();
	const res = await client.db(DATABASE).collection(col).findOne(query);

	if (res === null) return null;

	return fixListingFromMongo(res, dataGuard);
}

function fixListingForMongo(listing: Listing): ListingFromMongo {
	const { id, ...rest } = listing;
	return {
		_id: id,
		...rest,
	};
}

function fixListingFromMongo<T>(
	target: WithId<Document>,
	dataGuard: (target: unknown) => target is Listing<T>,
): Listing<T> {
	const { _id, ...rest } = target;
	const adjusted = { id: _id, ...rest };

	if (!dataGuard(adjusted)) {
		throw new Error("the returned document does not conform to the provided type");
	}

	return adjusted;
}
